# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.fluent.core.utils.fix_doc import fix_settings_doc


def test_fix_definition_list_in_settings_doc():
    old_doc = """Menu to define the rotor pitch and flapping angles.
 - blade-pitch-collective    : ,
 - blade-pitch-cyclic-sin    : ,
 - blade-pitch-cyclic-cos    : ,
For more details please consult the help option of the corresponding menu or TUI command."""
    new_doc = fix_settings_doc(old_doc)
    assert (
        new_doc
        == """Menu to define the rotor pitch and flapping angles.

 - blade-pitch-collective    : ,
 - blade-pitch-cyclic-sin    : ,
 - blade-pitch-cyclic-cos    : ,

For more details please consult the help option of the corresponding menu or TUI command."""
    )


def test_fix_wildcards_in_settings_doc():
    old_doc = """Read boundary profile data (*.prof, *.csv). Default is *.prof."""
    new_doc = fix_settings_doc(old_doc)
    assert (
        new_doc
        == r"""Read boundary profile data (\\*.prof, \\*.csv). Default is \\*.prof."""
    )
